class Node:
    def __init__(self, value):
        self.value = value
        self.left = None
        self.right = None


class TreeNode:

    def __init__(self):
        self.root = None

    def print(self):
        self.printDFS(self.root)

    def printDFS(self, root):
        if root is not None:
            print(root.value)
            self.printDFS(root.left)
            self.printDFS(root.right)

    def add(self, value):
        if self.root is None:
            self.root = Node(value)
        else:
            self._add(self.root, value)

    def _add(self, current, value):
        if current.value < value:
            if current.right is not None:
                self._add(current.right, value)
            else:
                current.right = Node(value)
        else:
            if current.left is not None:
                self._add(current.left, value)
            else:
                current.left = Node(value)

    def getHeight(self, current):
        if current is None:
            return 0
        return max(self.getHeight(current.left), self.getHeight(current.right)) + 1

    def isBalanced(self):
        return self._isBalanced(self.root)

    def _isBalanced(self, current):
        if current is None:
            return True
        heightDifference = abs(self.getHeight(current.left) - self.getHeight(current.right))
        if heightDifference > 1:
            return False
        return self._isBalanced(current.left) and self._isBalanced(current.right)


def makeTree():
    # values = [5, 4, 3, 2, 1, 7, 8]
    values = [5, 6, 4]
    treeNode = TreeNode()
    for v in values:
        treeNode.add(v)
    treeNode.print()
    print()
    print(treeNode.isBalanced())


# makeTree()


class Graph:

    def __init__(self, graph):
        self.graph = graph
        # print(self.graph)

    def add(self, A, B):
        if A not in self.graph:
            self.graph[A] = set()
        self.graph[A].add(B)

    def printer(self):
        print(self.graph)

    def getNeighbours(self, current):
        neighbours = self.graph[current]

    def DFS(self, start, target):
        stack = [start]
        visited = set()
        while stack:
            current = stack.pop()
            if current not in visited:
                visited.add(current)
                print(current)
                if current == target:
                    break
                for neighbour in self.graph[current]:
                    stack.append(neighbour)

    def BFS(self, start, target):
        queue = [start]
        visited = set()
        while queue:
            current = queue.pop(0)
            if current not in visited:
                visited.add(current)
                print(current)
                if current == target:
                    break
                for neighbour in self.graph[current]:
                    queue.append(neighbour)

    def getNeighbor(self, neighbors, graph):
        out = set()
        for neighbor in neighbors:
            out = out.union(graph[neighbor])
        return out

    def tierCount(self, start, target):
        graph = self.graph
        if target in graph[start]:
            return 1
        else:
            visited = set()
            queue = self.getNeighbor(graph[start], graph)
            count = 2
            while queue:
                if target in queue:
                    return count
                else:
                    count += 1
                    visited = visited.union(queue)
                    queue = self.getNeighbor(queue, graph) - visited
        return 0


# def runGraph():
#     graph = Graph({'0': {'1', '2'},
#                    '1': {'0', '3', '4'},
#                    '2': {'0'},
#                    '3': {'1'},
#                    '4': {'2', '3'}})
#     print(graph.tierCount('0', '3'))



import json
import urllib.request as urllib2
from distutils.version import StrictVersion


class PyPi(object):

    @classmethod
    def get_versions(self, package_name):
        url = "https://pypi.org/pypi/%s/json" % (package_name,)
        data = json.load(urllib2.urlopen(urllib2.Request(url)))
        versions = data["releases"].keys()
        versions = sorted(versions, key=StrictVersion)
        return versions

    @classmethod
    def get_current_version(self, package_name):
        return self.get_versions(package_name)[-1]

    @classmethod
    def get_latest_version(self, package_name):
        current = self.get_current_version(package_name).split(".")
        current[-1] = str(int(current[-1]) + 1)
        return ".".join(current)



class Stack():
    def __init__(self, size):
        self.stack = [None] * size
        self.top = -1

    def pop(self):
        if self.is_empty():
            raise Exception("Stack is empty!")
        saved = self.stack[self.top]
        self.top -= 1
        return saved

    def push(self, elem):
        if self.is_full():
            raise Exception("Stack is full!")
        self.top += 1
        self.stack[self.top] = elem

    def peek(self):
        return self.stack[self.top]

    def is_empty(self):
        return self.top == -1

    def is_full(self):
        return self.top == len(self.stack) - 1

    def __repr__(self):
        return str(self.stack)


# def main():
#     stack = Stack(5)
#     for i in range(1, 5):
#         stack.push(i)
#     print(stack)


# main()



import click
# from .visualsim import *


@click.command()
@click.option('--verbose', is_flag=True, help="Will print verbose messages.")
@click.option('--name', default='', help='Who are you?')
@click.option('--visualsim', default=False, help='CEG 4136 Lab Launcher Tool')
def cli(verbose, name, visualsim_bool):
    click.echo("Hello World. Welcome to PyNevin CLI")
    if verbose:
        click.echo("We are in the verbose mode.")
    if visualsim_bool:
        run()
    click.echo('Bye {0}'.format(name))



class Node:
    def __init__(self, value):
        self.value = value
        self.next = None

    def print(self):
        current = self
        string = ""
        while current is not None:
            string += str(current.value) + " "
            current = current.next
        print(string)


class LinkedList:
    def __init__(self):
        self.head = None

    def getHead(self):
        return self.head

    def add(self, value):
        if self.head is None:
            self.head = Node(value)
        else:
            current = self.head
            while current.next is not None:
                current = current.next
            current.next = Node(value)

    def remove_all_n(self, head, value):
        head, head.next = Node(0), head
        current = head
        while current.next:
            if current.next.value == value:
                current.next = current.next.next
            else:
                current = current.next
        return head.next

    def delete(self, value):
        current = self.head
        if current is None:
            return
        if current.value == value:
            self.head = current.next
            return
        previous = current
        current = current.next
        while current.next is not None:
            if current.value == value:
                previous.next = current.next
            previous = current
            current = current.next
        if current.value == value:
            previous.next = None
        else:
            raise Exception("Not found")

    def print(self):
        current = self.head
        string = ""
        while current is not None:
            string += str(current.value) + " "
            current = current.next
        print(string)

    def deleteDuplicates(self):
        current = self.head
        previous = None
        hashset = set()
        while current:
            if current.value in hashset:
                previous.next = current.next
            else:
                hashset.add(current.value)
                previous = current
            current = current.next

    def reverse(self):
        prev = None
        current = self.head
        while current is not None:
            next = current.next
            current.next = prev
            prev = current
            current = next
        self.head = prev

    def _reverse(self, current):
        if current is None:
            return
        if current.next is None:
            self.head = current
            return
        self._reverse(current.next)
        current.next.next = current
        current.next = None

    def reverse(self):
        self._reverse(self.head)
        return self.head


def scrambled(orig):
    dest = orig[:]
    import random
    random.shuffle(dest)
    return dest


def sumList(l1, l2):
    return _sumList(l1, l2, 0)


def _sumList(l1, l2, carry):
    if l1 is None and l2 is None and carry == 0:
        return None
    value = carry
    if l1 is not None:
        value += l1.value
    if l2 is not None:
        value += l2.value
    result = Node(value % 10)
    if l1 is not None or l2 is not None:
        v1 = None if l1 is None else l1.next
        v2 = None if l2 is None else l2.next
        carryBit = 1 if value >= 10 else 0
        more = _sumList(v1, v2, carryBit)
        result.next = more
    return result


# def LinkedListExample():
#     l1 = LinkedList()
#     l2 = LinkedList()
#     for val in [7, 1, 6]:
#         l1.add(val)
#     for val in [5, 9, 2]:
#         l2.add(val)
#     l1.print()
#     l2.print()


def findBeginning(head):
    slow = fast = head
    while fast is not None and fast.next is not None:
        slow = slow.next
        fast = fast.next.next
        if slow == fast:
            break
    if fast is None or fast.next is None:
        return None
    slow = head
    while slow != fast:
        slow = slow.next
        fast = fast.next
    return fast.value


# def makeLoop():
#     n1 = Node(1)
#     n2 = Node(2)
#     n3 = Node(3)
#     n4 = Node(4)
#     n5 = Node(5)
#     n1.next = n2
#     n2.next = n3
#     n3.next = n4
#     n4.next = n3
#     n5.next = n3
#     print(findBeginning(n2))



#  ============ 1.1 ============
def is_unique(s):
    if len(s) > 128:
        return False
    characters = [0] * 128
    for char in s:
        if characters[ord(char)]:
            return False
        characters[ord(char)] = True
    return True


#  ============ 1.2 ============

def get_required_letters(string):
    characters = [0] * 128
    for char in string:
        characters[ord(char)] += 1
    return characters


def is_permutation(s1, s2):
    return get_required_letters(s1) == get_required_letters(s2)


#  ============ 1.3 ============

def urlify(s, true_length):
    s, index = list(s), len(s)
    for i in reversed(range(true_length)):
        if s[i] == ' ':
            s[index - 1], s[index - 2], s[index - 3] = '0', '2', '%'
            index -= 3
        else:
            s[index - 1] = s[i]
            index -= 1
    return ''.join(s)


#  ============ 1.4 ============

def get_char_number(char):
    a, z, val = ord('a'), ord('z'), ord(char)
    return val - a if a <= val <= z else -1


def is_permutation_of_palindrome(phrase):
    odd_counter = 0
    table = [0] * (ord('z') - ord('a') + 1)
    for char in list(phrase):
        char_value = get_char_number(char)
        if char_value != -1:
            table[char_value] += 1
            if table[char_value] % 2 == 1:
                odd_counter += 1
            else:
                odd_counter -= 1
    return odd_counter <= 1


#  ============ 1.5 ============

def is_edit_away(first, second):
    if abs(len(first) - len(second)) > 1:
        return False
    if len(first) > len(second):
        first, second = second, first
    index1, index2 = 0, 0
    found_difference = False
    while index2 < len(second) and index1 < len(first):
        if first[index1] != second[index2]:
            if found_difference:
                return False
            else:
                found_difference = True
            if len(first) == len(second):
                index1 += 1
        else:
            index1 += 1
        index2 += 1
    return True


#  ============ 1.6 ============

def compress(string):
    compressed = ""
    count_consecutive = 0
    for i in range(len(string)):
        count_consecutive += 1
        if i + 1 >= len(string) or string[i] != string[i + 1]:
            compressed += string[i] + str(count_consecutive)
            count_consecutive = 0
    return compressed if len(compressed) < len(string) else string


#  ============ 1.7 ============

def rotate_matrix(matrix):
    n = len(matrix)
    if n == 0 or n != len(matrix[0]):
        return False
    for layer in range(n // 2):
        first = layer
        last = n - 1 - layer
        for i in range(last):
            offset = i - first
            top = matrix[first][i]
            matrix[first][i] = matrix[last - offset][first]
            matrix[last - offset][first] = matrix[last][last - offset]
            matrix[last][last - offset] = matrix[i][last]
            matrix[i][last] = top
    return True


#  ============ 1.8 ============

def zero_matrix(matrix):
    row_has_zero = False
    col_has_zero = False
    for j in range(len(matrix[0])):
        if matrix[0][j] == 0:
            row_has_zero = True
            break
    for i in range(len(matrix)):
        if matrix[i][0] == 0:
            col_has_zero = True
            break
    for i in range(len(matrix)):
        for j in range(len(matrix[0])):
            if matrix[i][j] == 0:
                matrix[i][0] = 0
                matrix[0][j] = 0

    def nullify_row(row):
        for _ in range(len(matrix[0])):
            matrix[row][_] = 0

    def nullify_col(col):
        for _ in range(len(matrix)):
            matrix[_][col] = 0

    for i in range(len(matrix)):
        if matrix[i][0] == 0:
            nullify_row(i)
    for j in range(len(matrix[0])):
        if matrix[0][j] == 0:
            nullify_col(j)
    if row_has_zero:
        nullify_row(0)
    if col_has_zero:
        nullify_col(0)
    return matrix


#  ============ 1.8 ============

def is_string_rotation(s1, s2):
    if len(s1) == len(s2) and len(s1) > 0:
        return s2 in s1 + s1
    return False


#  ============ Main ============

# if __name__ == '__main__':
#     print("Is Unique:", is_unique("abcdefghijklmnopqrstuvwxyz"))
#     print("Is Permutation:", is_permutation("god", "dog"))
#     print("URLify:", urlify("Mr John Smith    ", 13))
#     print("Is Permutation of Palindrome:",
#           is_permutation_of_palindrome("tact coa"))
#     print("Is One Edit Away:", is_edit_away("pale", "ple"))
#     print("String Compression:", compress("aabcccccaaa"))
#     matrix = [
#         [5, 0, 0, 0],
#         [0, 4, 0, 0],
#         [0, 0, 3, 0],
#         [0, 0, 0, 2],
#     ]
#     rotate_matrix(matrix)
#     print("Matrix Rotated:", matrix)
#
#     matrix = [
#         [1, 2, 3, 4],
#         [5, 0, 7, 8],
#         [9, 10, 11, 12],
#         [13, 14, 15, 16],
#     ]
#     print("Zero Matrix:", zero_matrix(matrix))
#
#     print("String Rotation:", is_string_rotation("waterbottle", "erbottlewat"))
